function [edges_3d,G1,G2] = Shearlet2Dedge(IMSS,n_levels)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% function [edges_3d,G1,G2] = Shearlet2Dedge(IMSS,n_levels)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PURPOSE: This is a 2D edge detection algorithm using
% shearlets.
% Input  IMSS is a 3D array containing the data:
%             intensity magnitudes at each voxel.
%        n_levels is the number of levels in the wavelet decomposition
%                 (default n_levels = 3)
% Output G1 and G2 are the estimated horizontal and vertical,
%             derivatives of the intensity at each voxel.
% Output edges_3d contains the estimated squared-magnitude of the
%             gradient at each voxel, normalized to have max 1.
% David A. Schug, Glenn R. Easley, Dianne P. O'Leary Fall 2012
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set the number of levels in the algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (nargin < 2)
    n_levels = 3;
end

% window_size is the size of the Meyer window

window_size = 17;
ascale = 2^(-3/4);
ascale_inv = 1/ascale;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generate smoothing, derivative and shearlet masks
% for convolution.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Smoothing operator:

v =  (1/16) * [ 1  2  1;
                2  4  2;
                1  2  1];

% Horizontal derivative:

d1  =  [ 1  0  -1;
         2  0  -2;
         1  0  -1];

d1 = d1 / sum(abs(d1(:)));

% Vertical derivative:

d2  =  [ 1   2   1;
         0   0   0;
        -1  -2  -1];

d2 = d2 / sum(abs(d2(:)));

%% Initialize the shearlet filters.

w_s = shearing_filters_Myer_odd(window_size, n_levels);

% Number of filters for shearing

n_f = size(w_s,3);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the edge detection algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Determine dimensions of image sequence.

[n_isr,n_isc,n_ist] = size(IMSS);

% Initialize horizontal and vertical derivatives. 

G1 = zeros(n_isr,n_isc,n_ist);
G2 = zeros(n_isr,n_isc,n_ist);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the loop through the image frames 1,..., n_ist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for  frame = 1 : n_ist
    
    % Estimate the horizontal and vertical derivatives
    % of the data using d1 and d2.
      
    a1 = symext_conv2D(IMSS(:,:,frame),d1);
    a2 = symext_conv2D(IMSS(:,:,frame),d2);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Begin the loop through the shearlet levels 1,...,n_levels.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    for  level = 1 : n_levels
        
        % Perform wavelet filtering.
        
        asmooth1 = symext_conv2D(a1,v);
        asmooth2 = symext_conv2D(a2,v);

        %%% SHEARING %%%
        
        % Perform the shearlet filtering.
        
        % Initialize storage for accumulated coefficients
        
        a1n = zeros(n_isr,n_isc);
        a2n = zeros(n_isr,n_isc);
        
        for  i_f = 1 : n_f,
            
            % Get shear filter for direction i_f.
            % (The fft's should be computed outside
            %  the loop for efficiency.)
            
            shear_f = real(fftshift(ifft2(fftshift(squeeze(w_s(:,:,i_f))))));
            
            % Filter the gradients.
            
            ashear1 = conv2(a1,shear_f,'same');
            ashear2 = conv2(a2,shear_f,'same');
            
            % Filter the smoothed gradients.
            
            ashearsmooth1 = conv2(asmooth1,shear_f,'same');
            ashearsmooth2 = conv2(asmooth2,shear_f,'same');
            
            % Find components of filtered gradients
            % that seem to be growing at the predicted rate.
            
            delta1 = (abs(ashear1) < ascale_inv*abs(ashearsmooth1)) & ...
                     (abs(ashear1) > ascale*abs(ashearsmooth1));
            delta2 = (abs(ashear2) < ascale_inv*abs(ashearsmooth2)) & ...
                     (abs(ashear2) > ascale*abs(ashearsmooth2));
            
            % Add retained components into accumulation. 
            
            a1n = a1n + ashear1 .* delta1;
            a2n = a2n + ashear2 .* delta2;
        
        end % for i_f = 1:n_f shearing loop

        %%% END SHEARING %%%

      % Add shearlet increments to a1 and a2.

        a1 = a1 + a1n;
        a2 = a2 + a2n;

    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % The loop through levels is complete.
    % Assign 2D shearlet edges to horizontal and vertical
    % derivatives 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    G1(:,:,frame) = a1;
    G2(:,:,frame) = a2;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The loop through frames is complete.
% Zero the boundary edge pixels.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

G1(1,:,:) = 0;
G1(:,1,:) = 0;
G1(:,:,1) = 0;

G2(1,:,:) = 0;
G2(:,1,:) = 0;
G2(:,:,1) = 0;

G1(end,:,:) = 0;
G1(:,end,:) = 0;
G1(:,:,end) = 0;

G2(end,:,:) = 0;
G2(:,end,:) = 0;
G2(:,:,end) = 0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the squared-magnitude of the gradient at each 
% voxel and return it in edges_3d.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

edges_3d = G1.^2 + G2.^2;

edges_3d = edges_3d / max(edges_3d(:));

