function [edges_3d,G1,G2,G3] = Shearlet3Dedge(IMSS,n_levels)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% function [edges_3d,G1,G2,G3] = Shearlet3Dedge(IMSS,n_levels)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PURPOSE: This is a 3D edge detection algorithm using shearlets.
% Input  IMSS is a 3D array containing the data: 
%             intensity magnitudes at each voxel.
%        n_levels is the number of levels in the wavelet decomposition
%                 (default n_levels = 3)
% Output G1, G2, and G3 are the estimated horizontal, vertical,
%             and time derivatives of the intensity at each voxel.
% Output edges_3d contains the estimated squared-magnitude of the
%             gradient at each voxel, normalized to have max 1.
% David Schug, Glenn R. Easley, Dianne P. O'Leary Fall 2012
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set the number of levels in the algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (nargin < 2)
    n_levels = 3;
end
%n_levels

% window_size is the size of the Meyer window
% In the current implementation, it must be odd.

window_size = 17;
ascale = 2^(-3/4);
ascale_inv = 1/ascale;

% 2^expd is the number of shearlet directions.

expd = 4;

% part_level determines the partitioning of the cube

part_level = 2;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generate smoothing, derivative, and shearlet masks
% for convolution. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Smoothing operator:

v= [1  2  1;
    2  4  2;
    1  2  1];

c= [1  1  1;
    1  2  1;
    1  1  1];

S3D(:,:,1) = c;
S3D(:,:,2) = v;
S3D(:,:,3) = c;
S3D = S3D / sum(abs(S3D(:)));

% Time derivative:                            

d3(:,:,1) = v;
d3(:,:,2) = zeros(3);
d3(:,:,3) = -v;
d3 = d3 / sum(abs(d3(:)));

% Horizontal derivative:                                

d1 = permute(d3,[2 3 1]);

% Vertical derivative:                                

d2 = permute(d3,[3 1 2]);

%% Initialize the shearlet filters.

Sh3D = shearing3D(window_size,part_level);

% Number of filters for shearing

n_f = size(Sh3D,4);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the edge detection algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Determine dimensions of image sequence
% and pad the time boundaries.

npad = 4;

[n_isr,n_isc,n_ist_in] = size(IMSS);

IMSSex(:,:,1)=squeeze(IMSS(:,:,1));
IMSSex(:,:,2)=squeeze(IMSS(:,:,1));

IMSSex(:,:,3:n_ist_in+2)=IMSS(:,:,1:n_ist_in);

IMSSex(:,:,n_ist_in+3)=squeeze(IMSS(:,:,n_ist_in));
IMSSex(:,:,n_ist_in+4)=squeeze(IMSS(:,:,n_ist_in));

IMSS=IMSSex;

% Estimate the horizontal, vertical, and time derivatives 
% of the data using d1, d2, and d3.           

G1 = conv3(IMSS,d1);
G2 = conv3(IMSS,d2);
G3 = conv3(IMSS,d3);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the loop through the shearlet levels 1,...,n_levels.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for level = 1 : n_levels

    % Perform wavelet filtering.

    G1s = conv3(G1,S3D);
    G2s = conv3(G2,S3D);
    G3s = conv3(G3,S3D);   

    %%% SHEARING %%%

    % Perform the shearlet filtering.

    % Initialize storage for accumulated coefficients

    G1n = zeros(n_isr,n_isc,n_ist_in+npad);
    G2n = zeros(n_isr,n_isc,n_ist_in+npad);
    G3n = zeros(n_isr,n_isc,n_ist_in+npad);

    for i_f = 1:n_f

       % Get shear filter for direction i_f.
       % (The fft's should be computed outside
       %  the loop for efficiency.)

       shear_f3D = real(fftshift(ifftn(fftshift(Sh3D(:,:,:,i_f)))));

       % Filter the gradients.

       ashear1 = conv3(G1,shear_f3D);
       ashear2 = conv3(G2,shear_f3D);
       ashear3 = conv3(G3,shear_f3D);

       % Filter the smoothed gradients.

       ashearsmooth1 = conv3(G1s,shear_f3D);
       ashearsmooth2 = conv3(G2s,shear_f3D);
       ashearsmooth3 = conv3(G3s,shear_f3D);

       % Retain components of filtered smoothed gradients
       % greater in magnitude than filtered gradients.

       delta1 = (abs(ashear1) < ascale_inv*abs(ashearsmooth1)) & ...
                (abs(ashear1) > ascale*abs(ashearsmooth1));
       delta2 = (abs(ashear2) < ascale_inv*abs(ashearsmooth2)) & ...
                (abs(ashear2) > ascale*abs(ashearsmooth2));
       delta3 = (abs(ashear3) < ascale_inv*abs(ashearsmooth3)) & ...
                (abs(ashear3) > ascale*abs(ashearsmooth3));

       % Add retained components into G1n, G2n, and G3n.

       G1n = G1n + ashear1 .* delta1;
       G2n = G2n + ashear2 .* delta2;
       G3n = G3n + ashear3 .* delta3;

    end  % for i_f = 1:n_f shearing loop

    %%% END SHEARING %%%

    % Add shearlet increments to G1, G2, and G3.

    G1 = G1 + G1n;
    G2 = G2 + G2n;
    G3 = G3 + G3n; 

end  % for level = 1 : n_levels           

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The loop through levels is complete.
% Trim the padded horizontal, vertical, and time 
% derivatives.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

G1 = G1(:,:,3:n_ist_in+2);
G2 = G2(:,:,3:n_ist_in+2);
G3 = G3(:,:,3:n_ist_in+2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Zero the boundary edge pixels.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

G1(1,:,:) = 0;
G1(:,1,:) = 0;
G1(:,:,1) = 0;

G2(1,:,:) = 0;
G2(:,1,:) = 0;
G2(:,:,1) = 0;

G3(1,:,:) = 0;
G3(:,1,:) = 0;
G3(:,:,1) = 0;

G1(end,:,:) = 0;
G1(:,end,:) = 0;
G1(:,:,end) = 0;

G2(end,:,:) = 0;
G2(:,end,:) = 0;
G2(:,:,end) = 0;

G3(end,:,:) = 0;
G3(:,end,:) = 0;
G3(:,:,end) = 0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the squared-magnitude of the gradient at each 
% voxel and return it in edges_3d.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

edges_3d = G1.^2 + G2.^2 + G3.^2;

edges_3d = edges_3d / max(edges_3d(:));

