function [edges_3d,G1,G2] = Wavelet2Dedge(IMSS,n_levels)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% function [edges_3d,G1,G2] = Wavelet2Dedge(IMSS,n_levels)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PURPOSE: This is a 2D edge detection algorithm using wavelets.
% Input  IMSS is a 3D array containing the data:
%        intensity magnitudes at each voxel.
%        n_levels is the number of levels in the wavelet decomposition
%                 (default n_levels = 3)
% Output G1 and G2 are the estimated horizontal and vertical
%        derivatives of the intensity at each voxel.
% Output edges_3d contains the estimated squared-magnitude of the
%             gradient at each voxel, normalized to max 1.
% David A. Schug, Glenn R. Easley, Dianne P. O'Leary Fall 2012
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set the number of levels in the algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if (nargin < 2)
    n_levels = 3;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generate smoothing and derivative masks
% for convolution.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Smoothing operator:

v =  (1/16) * [ 1  2  1;
                2  4  2;
                1  2  1];

% Horizontal derivative:

d1  = (1/8) * [ 1  0  -1;
                2  0  -2;
                1  0  -1];

% Vertical derivative:

d2  = (1/8) *  [ 1   2   1;
                 0   0   0;
                -1  -2  -1];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the edge detection algorithm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Determine dimensions of image sequence.

[n_isr,n_isc,n_ist] = size(IMSS);

% Initialize horizontal and vertical derivatives.

G1 = zeros(n_isr,n_isc,n_ist);
G2 = zeros(n_isr,n_isc,n_ist);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Begin the loop through the image frames 1,..., n_ist.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for frame = 1: n_ist
    
    % Estimate the horizontal and vertical derivatives
    % of the data using d1 and d2.
    
    a1 = symext_conv2D(IMSS(:,:,frame),d1);
    a2 = symext_conv2D(IMSS(:,:,frame),d2);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Begin the loop through the wavelet levels 1,...,s.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    for level = 1 : n_levels
        
        % Perform wavelet filtering.
        
        asmooth1 = symext_conv2D(a1,v);
        asmooth2 = symext_conv2D(a2,v);
        
        % Add wavelet increments to a1 and a2.
        
        a1 = a1 .* ( abs(a1) <= abs(asmooth1)) + asmooth1 .* (abs(a1) > abs(asmooth1) );
        a2 = a2 .* ( abs(a2) <= abs(asmooth2)) + asmooth2 .* (abs(a2) > abs(asmooth2) );

    end   % for level = 1 : n_levels
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % The loop through levels is complete.
    % Assign 2D wavelet component edges to horizontal
    % and vertical derivatives
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    G1(:,:,frame) = a1;
    G2(:,:,frame) = a2;
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The loop through frames is complete.
% Zero the boundary edge pixels.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

G1(1,:,:) = 0;
G1(:,1,:) = 0;
G1(:,:,1) = 0;

G2(1,:,:) = 0;
G2(:,1,:) = 0;
G2(:,:,1) = 0;

G1(end,:,:) = 0;
G1(:,end,:) = 0;
G1(:,:,end) = 0;

G2(end,:,:) = 0;
G2(:,end,:) = 0;
G2(:,:,end) = 0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the squared-magnitude of the gradient at each 
% voxel and return it in edges_3d.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

edges_3d = G1.^2 + G2.^2;

edges_3d = edges_3d / max(edges_3d(:));
